import pygad
import random
import numpy

num_generations = 1

initial_population = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]

#### Define the fitness functions in the top-level of the module so that they are picklable and usable in the process-based parallel processing works.
#### If the functions are defined inside a class/method/function, they are not picklable and this error is raised: AttributeError: Can't pickle local object
#### Process-based parallel processing must have the used functions picklable.
def fitness_func_no_batch_single(ga, solution, idx):
    return random.random()

def fitness_func_batch_single(ga, soluions, idxs):
    return numpy.random.uniform(size=len(soluions))

def fitness_func_no_batch_multi(ga, solution, idx):
    return [random.random(), random.random()]

def fitness_func_batch_multi(ga, soluions, idxs):
    f = []
    for sol in soluions:
        f.append([random.random(), random.random()])
    return f

def output_adaptive_mutation(gene_space=None,
                             gene_type=float,
                             num_genes=10,
                             mutation_by_replacement=False,
                             random_mutation_min_val=-1,
                             random_mutation_max_val=1,
                             init_range_low=-4,
                             init_range_high=4,
                             initial_population=None,
                             mutation_probability=None,
                             mutation_num_genes=None,
                             fitness_batch_size=None,
                             mutation_type="adaptive",
                             parent_selection_type='sss',
                             parallel_processing=None,
                             multi_objective=False):

    if fitness_batch_size is None or (type(fitness_batch_size) in pygad.GA.supported_int_types and fitness_batch_size == 1):
        if multi_objective == True:
            fitness_func = fitness_func_no_batch_multi
        else:
            fitness_func = fitness_func_no_batch_single
    elif (type(fitness_batch_size) in pygad.GA.supported_int_types and fitness_batch_size > 1):
        if multi_objective == True:
            fitness_func = fitness_func_batch_multi
        else:
            fitness_func = fitness_func_batch_single
    
    ga_instance = pygad.GA(num_generations=num_generations,
                           num_parents_mating=5,
                           fitness_func=fitness_func,
                           sol_per_pop=10,
                           num_genes=num_genes,
                           gene_space=gene_space,
                           gene_type=gene_type,
                           initial_population=initial_population,
                           init_range_low=init_range_low,
                           init_range_high=init_range_high,
                           parent_selection_type=parent_selection_type,
                           random_mutation_min_val=random_mutation_min_val,
                           random_mutation_max_val=random_mutation_max_val,
                           allow_duplicate_genes=True,
                           mutation_by_replacement=mutation_by_replacement,
                           save_solutions=True,
                           ## Use a static 'mutation_probability'.
                           ## An ambigius error in GitHub actions happen when using mutation_num_genes and mutation_probability. I do not know the reason.
                           # mutation_num_genes=mutation_num_genes,
                           mutation_probability=[0.2, 0.1],
                           mutation_type=mutation_type,
                           suppress_warnings=True,
                           fitness_batch_size=fitness_batch_size,
                           parallel_processing=parallel_processing,
                           random_seed=1)

    ga_instance.run()

    return None, ga_instance

def test_adaptive_mutation():
    result, ga_instance = output_adaptive_mutation()

    # assert result == True

def test_adaptive_mutation_gene_space():
    result, ga_instance = output_adaptive_mutation(gene_space=range(10))

    # assert result == True

def test_adaptive_mutation_int_gene_type():
    result, ga_instance = output_adaptive_mutation(gene_type=int)

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type():
    result, ga_instance = output_adaptive_mutation(gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2])

    # assert result == True

def test_adaptive_mutation_nested_gene_space():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]])
    # assert result == True

def test_adaptive_mutation_nested_gene_type():
    result, ga_instance = output_adaptive_mutation(gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]])

    # assert result == True

def test_adaptive_mutation_initial_population():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population)

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=1)

def test_adaptive_mutation_fitness_batch_size_2():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=2)

def test_adaptive_mutation_fitness_batch_size_3():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=3)

def test_adaptive_mutation_fitness_batch_size_4():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=4)

def test_adaptive_mutation_fitness_batch_size_5():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=5)

def test_adaptive_mutation_fitness_batch_size_6():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=6)

def test_adaptive_mutation_fitness_batch_size_7():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=7)

def test_adaptive_mutation_fitness_batch_size_8():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=8)

def test_adaptive_mutation_fitness_batch_size_9():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=9)

def test_adaptive_mutation_fitness_batch_size_10():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=10)

#### Single-Objective Mutation Probability
def test_adaptive_mutation_mutation_probability():
    result, ga_instance = output_adaptive_mutation(mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_probability():
    result, ga_instance = output_adaptive_mutation(gene_space=range(10),
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_probability():
    result, ga_instance = output_adaptive_mutation(gene_type=int,
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_probability():
    result, ga_instance = output_adaptive_mutation(gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_probability():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_probability=[0.2, 0.1])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_probability():
    result, ga_instance = output_adaptive_mutation(gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_probability():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_probability():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=1,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_2_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=2,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_3_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=3,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_4_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=4,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_5_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=5,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_6_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=6,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_7_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=7,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_8_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=8,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_9_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=9,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_10_mutation_probability():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=10,
                                                   mutation_probability=[0.2, 0.1])


#### Single-Objective Mutation Number of Genes
def test_adaptive_mutation_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(gene_space=range(10),
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(gene_type=int,
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_num_genes=[6, 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_num_genes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=1,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=2,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=3,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=4,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=5,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=6,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=7,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=8,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=9,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=10,
                                                   mutation_num_genes=[6, 4])

#### Multi-Objective Mutation Probability
def test_adaptive_mutation_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=range(10),
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=int,
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_probability=[0.2, 0.1])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_probability_multi_objective():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_multi_objective():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=1,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_2_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=2,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_3_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=3,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_4_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=4,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_5_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=5,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_6_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=6,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_7_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=7,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_8_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=8,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_9_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=9,
                                                   mutation_probability=[0.2, 0.1])

def test_adaptive_mutation_fitness_batch_size_10_mutation_probability_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=10,
                                                   mutation_probability=[0.2, 0.1])

#### Multi-Objective Mutation Number of Genes
def test_adaptive_mutation_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=range(10),
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=int,
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_num_genes=[6, 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_num_genes_multi_objective():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes_multi_objective():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=1,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=2,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=3,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=4,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=5,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=6,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=7,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=8,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=9,
                                                   mutation_num_genes=[6, 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes_multi_objective():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,fitness_batch_size=10,
                                                   mutation_num_genes=[6, 4])

######## Parallel Processing
#### #### Threads

#### Single-Objective Mutation Probability
def test_adaptive_mutation_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_space=range(10),
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_type=int,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_probability_parallel_processing_threads():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_parallel_processing_threads():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=1,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=2,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=3,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=4,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=5,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=6,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=7,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=8,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=9,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_probability_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=10,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])


#### Single-Objective Mutation Number of Genes
def test_adaptive_mutation_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_space=range(10),
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_type=int,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_num_genes_parallel_processing_threads():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes_parallel_processing_threads():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=1,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=2,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=3,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=4,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=5,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=6,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=7,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=8,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=9,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=10,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

#### Multi-Objective Mutation Probability
def test_adaptive_mutation_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=range(10),
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=int,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_probability_multi_objective_parallel_processing_threads():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_multi_objective_parallel_processing_threads():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=1,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=2,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=3,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=4,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=5,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=6,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=7,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=8,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=9,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_probability_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=10,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['thread', 4])

#### Multi-Objective Mutation Number of Genes
def test_adaptive_mutation_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=range(10),
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=int,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_num_genes_multi_objective_parallel_processing_threads():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes_multi_objective_parallel_processing_threads():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=1,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=2,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=3,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=4,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=5,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=6,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=7,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=8,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=9,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes_multi_objective_parallel_processing_threads():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,fitness_batch_size=10,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['thread', 4])


#### #### Processes

#### Single-Objective Mutation Probability
def test_adaptive_mutation_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_space=range(10),
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_type=int,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_probability_parallel_processing_processes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_parallel_processing_processes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=1,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=2,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=3,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=4,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=5,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=6,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=7,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=8,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=9,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_probability_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=10,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])


#### Single-Objective Mutation Number of Genes
def test_adaptive_mutation_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_space=range(10),
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_type=int,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_num_genes_parallel_processing_processes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes_parallel_processing_processes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=1,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=2,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=3,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=4,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=5,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=6,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=7,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=8,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=9,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(fitness_batch_size=10,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

#### Multi-Objective Mutation Probability
def test_adaptive_mutation_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=range(10),
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=int,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_probability_multi_objective_parallel_processing_processes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_multi_objective_parallel_processing_processes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=1,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=2,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=3,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=4,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=5,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=6,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=7,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=8,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=9,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_probability_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=10,
                                                   mutation_probability=[0.2, 0.1], 
                                                   parallel_processing=['process', 4])

#### Multi-Objective Mutation Number of Genes
def test_adaptive_mutation_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=range(10),
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_int_gene_type_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=int,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_gene_space_gene_type_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space={"low": 0, "high": 10},
                                                   gene_type=[float, 2],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])
    # assert result == True

def test_adaptive_mutation_nested_gene_type_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_nested_gene_space_nested_gene_type_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   gene_space=[[0, 1, 2, 3, 4], 
                                                                numpy.arange(5, 10), 
                                                                range(10, 15),
                                                                {"low": 15, "high": 20},
                                                                {"low": 20, "high": 30, "step": 2},
                                                                None,
                                                                numpy.arange(30, 35),
                                                                numpy.arange(35, 40),
                                                                numpy.arange(40, 45),
                                                                [45, 46, 47, 48, 49]],
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_mutation_num_genes_multi_objective_parallel_processing_processes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes_multi_objective_parallel_processing_processes():
    global initial_population
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   initial_population=initial_population,
                                                   gene_type=[int, float, numpy.float64, [float, 3], [float, 4], numpy.int16, [numpy.float32, 1], int, float, [float, 3]],
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

    # assert result == True

def test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=1,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=2,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=3,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=4,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=5,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=6,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=7,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=8,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,
                                                   fitness_batch_size=9,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])

def test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes_multi_objective_parallel_processing_processes():
    result, ga_instance = output_adaptive_mutation(multi_objective=True,fitness_batch_size=10,
                                                   mutation_num_genes=[6, 4], 
                                                   parallel_processing=['process', 4])


if __name__ == "__main__":
    #### Single-objective mutation_probability
    print()
    test_adaptive_mutation_mutation_probability()
    print()

    test_adaptive_mutation_int_gene_type_mutation_probability()
    print()

    test_adaptive_mutation_gene_space_mutation_probability()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_probability()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_probability()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_probability()
    print()

    test_adaptive_mutation_initial_population_mutation_probability()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_probability()
    print()

    #### Single-objective mutation_num_genes
    print()
    test_adaptive_mutation_mutation_num_genes()
    print()

    test_adaptive_mutation_int_gene_type_mutation_num_genes()
    print()

    test_adaptive_mutation_gene_space_mutation_num_genes()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_num_genes()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_num_genes()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_num_genes()
    print()

    test_adaptive_mutation_initial_population_mutation_num_genes()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes()
    print()

    #### Multi-objective mutation_probability
    print()
    test_adaptive_mutation_mutation_probability()
    print()

    test_adaptive_mutation_int_gene_type_mutation_probability()
    print()

    test_adaptive_mutation_gene_space_mutation_probability()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_probability()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_probability()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_probability()
    print()

    test_adaptive_mutation_initial_population_mutation_probability()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_probability()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_probability()
    print()



    ######## Parallel Processing
    #### #### Threads
    #### Single-objective mutation_probability
    print()
    test_adaptive_mutation_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_int_gene_type_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_gene_space_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_initial_population_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_probability_parallel_processing_threads()
    print()

    #### Single-objective mutation_num_genes
    print()
    test_adaptive_mutation_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_int_gene_type_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_gene_space_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_initial_population_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes_parallel_processing_threads()
    print()

    #### Multi-objective mutation_probability
    print()
    test_adaptive_mutation_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_int_gene_type_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_gene_space_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_initial_population_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_probability_parallel_processing_threads()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_probability_parallel_processing_threads()
    print()

    #### #### Processes
    #### Single-objective mutation_probability
    print()
    test_adaptive_mutation_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_int_gene_type_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_gene_space_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_initial_population_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_probability_parallel_processing_processes()
    print()

    #### Single-objective mutation_num_genes
    print()
    test_adaptive_mutation_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_int_gene_type_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_gene_space_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_initial_population_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_num_genes_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_num_genes_parallel_processing_processes()
    print()

    #### Multi-objective mutation_probability
    print()
    test_adaptive_mutation_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_int_gene_type_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_gene_space_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_gene_space_gene_type_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_nested_gene_space_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_nested_gene_type_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_initial_population_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_initial_population_nested_gene_type_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_1_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_2_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_3_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_4_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_5_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_6_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_7_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_8_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_9_mutation_probability_parallel_processing_processes()
    print()

    test_adaptive_mutation_fitness_batch_size_10_mutation_probability_parallel_processing_processes()
    print()





import pygad
import random
import numpy

num_generations = 5

initial_population = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]

def validate_gene_type_and_rounding(gene_space=None,
                                    gene_type=float,
                                    num_genes=10,
                                    mutation_by_replacement=False,
                                    random_mutation_min_val=-1,
                                    random_mutation_max_val=1,
                                    init_range_low=-4,
                                    init_range_high=4,
                                    initial_population=None,
                                    crossover_probability=None,
                                    mutation_probability=None,
                                    crossover_type=None,
                                    mutation_type=None,
                                    gene_constraint=None,
                                    parent_selection_type='sss',
                                    multi_objective=False):

    def fitness_func_no_batch_single(ga, solution, idx):
        return random.random()

    def fitness_func_no_batch_multi(ga, solution, idx):
        return [random.random(), random.random()]

    if multi_objective == True:
        fitness_func = fitness_func_no_batch_multi
    else:
        fitness_func = fitness_func_no_batch_single

    ga_instance = pygad.GA(num_generations=num_generations,
                           num_parents_mating=5,
                           fitness_func=fitness_func,
                           sol_per_pop=10,
                           num_genes=num_genes,
                           gene_space=gene_space,
                           gene_constraint=gene_constraint,
                           gene_type=gene_type,
                           parent_selection_type=parent_selection_type,
                           initial_population=initial_population,
                           init_range_low=init_range_low,
                           init_range_high=init_range_high,
                           random_mutation_min_val=random_mutation_min_val,
                           random_mutation_max_val=random_mutation_max_val,
                           allow_duplicate_genes=True,
                           mutation_by_replacement=mutation_by_replacement,
                           save_solutions=True,
                           crossover_probability=crossover_probability,
                           mutation_probability=mutation_probability,
                           crossover_type=crossover_type,
                           mutation_type=mutation_type,
                           suppress_warnings=True,
                           random_seed=1)

    ga_instance.run()

    for sol_idx in range(len(ga_instance.solutions)):
        for gene_idx in range(ga_instance.num_genes):
            if ga_instance.gene_type_single:
                dtype = ga_instance.gene_type
            else:
                dtype = ga_instance.gene_type[gene_idx]

            if dtype[0] is float:
                # NumPy converts the Python float data type to numpy.float64. Both are identical.
                assert type(ga_instance.solutions[sol_idx][gene_idx]) in [dtype[0], numpy.float64]
            else:
                assert type(ga_instance.solutions[sol_idx][gene_idx]) is dtype[0]

            if dtype[1] is None:
                pass
            else:
                num_decimals = len(str(ga_instance.solutions[sol_idx][gene_idx]).split('.')[1])
                # The numbers may not have the exact precision.
                # For example, the float number might be 5.7, and we set the precision to 2.
                # Because there is no enough digits after the decimal point, we cannot meet the precision of 2.
                # We only care about not exceeding the user-defined precision.
                assert num_decimals <= dtype[1]

    return ga_instance

def test_nested_gene_type():
    ga_instance = validate_gene_type_and_rounding(gene_type=[numpy.int32,
                                                             numpy.float16,
                                                             numpy.float32,
                                                             [numpy.float16, 3],
                                                             [numpy.float32, 4],
                                                             numpy.int16,
                                                             [numpy.float32, 1],
                                                             numpy.int32,
                                                             numpy.float16,
                                                             numpy.float64])

def test_single_gene_type_float16():
    ga_instance = validate_gene_type_and_rounding(gene_type=[numpy.float16, 2])

def test_single_gene_type_int32():
    ga_instance = validate_gene_type_and_rounding(gene_type=numpy.int32)

def test_single_gene_space_single_gene_type():
    ga_instance = validate_gene_type_and_rounding(gene_space={"low": 0, "high": 10},
                                                  gene_type=[float, 2])

def test_nested_gene_space_single_gene_type():
    ga_instance = validate_gene_type_and_rounding(gene_space=[[0, 1, 2, 3, 4],
                                                              numpy.arange(5, 10),
                                                              range(10, 15),
                                                              {"low": 15, "high": 20},
                                                              {"low": 20, "high": 30, "step": 2},
                                                              None,
                                                              numpy.arange(30, 35),
                                                              numpy.arange(35, 40),
                                                              numpy.arange(40, 45),
                                                              [45, 46, 47, 48, 49]],
                                                  gene_type=[numpy.float16, 1])

def test_nested_gene_space_nested_gene_type():
    ga_instance = validate_gene_type_and_rounding(gene_space=[[0, 1, 2, 3, 4],
                                                              numpy.arange(5, 10),
                                                              range(10, 15),
                                                              {"low": 15, "high": 20},
                                                              {"low": 20, "high": 30, "step": 2},
                                                              None,
                                                              numpy.arange(30, 35),
                                                              numpy.arange(35, 40),
                                                              numpy.arange(40, 45),
                                                              [45, 46, 47, 48, 49]],
                                                  gene_type=[int,
                                                             float,
                                                             numpy.float64,
                                                             [float, 3],
                                                             [float, 4],
                                                             numpy.int16,
                                                             [numpy.float32, 1],
                                                             int,
                                                             float,
                                                             [float, 3]])

def test_single_gene_space_nested_gene_type():
    ga_instance = validate_gene_type_and_rounding(gene_space=numpy.arange(0, 100),
                                                  gene_type=[int,
                                                             float,
                                                             numpy.float64,
                                                             [float, 3],
                                                             [float, 4],
                                                             numpy.int16,
                                                             [numpy.float32, 1],
                                                             int,
                                                             float,
                                                             [float, 3]])

def test_custom_initial_population_single_gene_type():
    global initial_population
    ga_instance = validate_gene_type_and_rounding(initial_population=initial_population,
                                                  gene_type=[numpy.float16, 2])

def test_custom_initial_population_nested_gene_type():
    global initial_population
    ga_instance = validate_gene_type_and_rounding(initial_population=initial_population,
                                                  gene_type=[int,
                                                             float,
                                                             numpy.float64,
                                                             [float, 3],
                                                             [float, 4],
                                                             numpy.int16,
                                                             [numpy.float32, 1],
                                                             int,
                                                             float,
                                                             [float, 3]])

if __name__ == "__main__":
    #### Single-objective
    print()
    test_nested_gene_type()
    print()

    test_single_gene_type_float16()
    print()

    test_single_gene_type_int32()
    print()

    test_single_gene_space_single_gene_type()
    print()

    test_nested_gene_space_single_gene_type()
    print()

    test_nested_gene_space_nested_gene_type()
    print()

    test_single_gene_space_nested_gene_type()
    print()

    test_custom_initial_population_single_gene_type()
    print()

    test_custom_initial_population_nested_gene_type()
    print()



import numpy
import pygad

class NSGA2:

    def __init__(self):
        pass

    def get_non_dominated_set(self, curr_solutions):
        """
        Get the set of non-dominated solutions from the current set of solutions.
    
        Parameters
        ----------
        curr_solutions : TYPE
            The set of solutions to find its non-dominated set.
    
        Returns
        -------
        dominated_set : TYPE
            A set of the dominated solutions.
        non_dominated_set : TYPE
            A set of the non-dominated set.
    
        """
        # List of the members of the current dominated pareto front/set.
        dominated_set = []
        # List of the non-members of the current dominated pareto front/set.
        # The non-dominated set is the pareto front set.
        non_dominated_set = []
        for idx1, sol1 in enumerate(curr_solutions):
            # Flag indicates whether the solution is a member of the current dominated set.
            is_not_dominated = True
            for idx2, sol2 in enumerate(curr_solutions):
                if idx1 == idx2:
                    continue

                # Zipping the 2 solutions so the corresponding genes are in the same list.
                # The returned array is of size (N, 2) where N is the number of genes.
                two_solutions = numpy.array(list(zip(sol1[1], sol2[1])))

                # Use < for minimization problems and > for maximization problems.
                # Checking if any solution dominates the current solution by applying the 2 conditions.
                # gr_eq (greater than or equal): All elements must be True.
                # gr (greater than): Only 1 element must be True.
                gr_eq = two_solutions[:, 1] >= two_solutions[:, 0]
                gr = two_solutions[:, 1] > two_solutions[:, 0]

                # If the 2 conditions hold, then a solution (sol2) dominates the current solution (sol1).
                # The current solution (sol1) is not considered a member of the non-dominated set.
                if gr_eq.all() and gr.any():
                    # Set the is_not_dominated flag to False because another solution dominates the current solution (sol1)
                    is_not_dominated = False
                    # DO NOT insert the current solution in the current non-dominated set.
                    # Instead, insert it into the dominated set.
                    dominated_set.append(sol1)
                    break
                else:
                    # Reaching here means the solution does not dominate the current solution.
                    pass
    
            # If the flag is True, then no solution dominates the current solution.
            # Insert the current solution (sol1) into the non-dominated set.
            if is_not_dominated:
                non_dominated_set.append(sol1)
    
        # Return the dominated and non-dominated sets.
        return dominated_set, non_dominated_set

    def non_dominated_sorting(self, fitness):
        """
        Apply non-dominant sorting over the fitness to create the pareto fronts based on non-dominated sorting of the solutions.
    
        Parameters
        ----------
        fitness : TYPE
            An array of the population fitness across all objective function.
    
        Returns
        -------
        pareto_fronts : TYPE
            An array of the pareto fronts.

        """

        # Verify that the problem is multi-objective optimization as non-dominated sorting is only applied to multi-objective problems.
        if type(fitness[0]) in [list, tuple, numpy.ndarray]:
            pass
        elif type(fitness[0]) in self.supported_int_float_types:
            raise TypeError('Non-dominated sorting is only applied when optimizing multi-objective problems.\n\nBut a single-objective optimization problem found as the fitness function returns a single numeric value.\n\nTo use multi-objective optimization, consider returning an iterable of any of these data types:\n1)list\n2)tuple\n3)numpy.ndarray')
        else:
            raise TypeError(f'Non-dominated sorting is only applied when optimizing multi-objective problems. \n\nTo use multi-objective optimization, consider returning an iterable of any of these data types:\n1)list\n2)tuple\n3)numpy.ndarray\n\nBut the data type {type(fitness[0])} found.')

        # A list of all non-dominated sets.
        pareto_fronts = []
    
        # The remaining set to be explored for non-dominance.
        # Initially it is set to the entire population.
        # The solutions of each non-dominated set are removed after each iteration.
        remaining_set = fitness.copy()
    
        # Zipping the solution index with the solution's fitness.
        # This helps to easily identify the index of each solution.
        # Each element has:
            # 1) The index of the solution.
            # 2) An array of the fitness values of this solution across all objectives.
        remaining_set = list(zip(range(0, fitness.shape[0]), remaining_set))
    
        # A list mapping the index of each pareto front to the set of solutions in this front.
        solutions_fronts_indices = [-1]*len(remaining_set)
        solutions_fronts_indices = numpy.array(solutions_fronts_indices)
    
        # Index of the current pareto front.
        front_index = -1
        while len(remaining_set) > 0:
            front_index += 1

            # Get the current non-dominated set of solutions.
            remaining_set, pareto_front = self.get_non_dominated_set(curr_solutions=remaining_set)
            pareto_front = numpy.array(pareto_front, dtype=object)
            pareto_fronts.append(pareto_front)
    
            solutions_indices = pareto_front[:, 0].astype(int)
            solutions_fronts_indices[solutions_indices] = front_index

        return pareto_fronts, solutions_fronts_indices

    def crowding_distance(self, pareto_front, fitness):
        """
        Calculate the crowding distance for all solutions in the current pareto front.
    
        Parameters
        ----------
        pareto_front : TYPE
            The set of solutions in the current pareto front.
        fitness : TYPE
            The fitness of the current population.
    
        Returns
        -------
        obj_crowding_dist_list : TYPE
            A nested list of the values for all objectives alongside their crowding distance.
        crowding_dist_sum : TYPE
            A list of the sum of crowding distances across all objectives for each solution.
        crowding_dist_front_sorted_indices : TYPE
            The indices of the solutions (relative to the current front) sorted by the crowding distance.
        crowding_dist_pop_sorted_indices : TYPE
            The indices of the solutions (relative to the population) sorted by the crowding distance.
        """
    
        # Each solution in the pareto front has 2 elements:
            # 1) The index of the solution in the population.
            # 2) A list of the fitness values for all objectives of the solution.
        # Before proceeding, remove the indices from each solution in the pareto front.
        pareto_front_no_indices = numpy.array([pareto_front[:, 1][idx] for idx in range(pareto_front.shape[0])])
    
        # If there is only 1 solution, then return empty arrays for the crowding distance.
        if pareto_front_no_indices.shape[0] == 1:
            # There is only 1 index.
            return numpy.array([]), numpy.array([]), numpy.array([0]), pareto_front[:, 0].astype(int)
    
        # An empty list holding info about the objectives of each solution. The info includes the objective value and crowding distance.
        obj_crowding_dist_list = []
        # Loop through the objectives to calculate the crowding distance of each solution across all objectives.
        for obj_idx in range(pareto_front_no_indices.shape[1]):
            obj = pareto_front_no_indices[:, obj_idx]
            # This variable has a nested list where each child list zip the following together:
                # 1) The index of the objective value.
                # 2) The objective value.
                # 3) Initialize the crowding distance by zero.
            obj = list(zip(range(len(obj)), obj, [0]*len(obj)))
            obj = [list(element) for element in obj]
            # This variable is the sorted version where sorting is done by the objective value (second element).
            # Note that the first element is still the original objective index before sorting.
            obj_sorted = sorted(obj, key=lambda x: x[1])
        
            # Get the minimum and maximum values for the current objective.
            obj_min_val = min(fitness[:, obj_idx])
            obj_max_val = max(fitness[:, obj_idx])
            denominator = obj_max_val - obj_min_val
            # To avoid division by zero, set the denominator to a tiny value.
            if denominator == 0:
                denominator = 0.0000001
    
            # Set the crowding distance to the first and last solutions (after being sorted) to infinity.
            inf_val = float('inf')
            # crowding_distance[0] = inf_val
            obj_sorted[0][2] = inf_val
            # crowding_distance[-1] = inf_val
            obj_sorted[-1][2] = inf_val
    
            # If there are only 2 solutions in the current pareto front, then do not proceed.
            # The crowding distance for such 2 solutions is infinity.
            if len(obj_sorted) <= 2:
                obj_crowding_dist_list.append(obj_sorted)
                break
    
            for idx in range(1, len(obj_sorted)-1):
                # Calculate the crowding distance.
                crowding_dist = obj_sorted[idx+1][1] - obj_sorted[idx-1][1]
                crowding_dist = crowding_dist / denominator
                # Insert the crowding distance back into the list to override the initial zero.
                obj_sorted[idx][2] = crowding_dist
        
            # Sort the objective by the original index at index 0 of each child list.
            obj_sorted = sorted(obj_sorted, key=lambda x: x[0])
            obj_crowding_dist_list.append(obj_sorted)
    
        obj_crowding_dist_list = numpy.array(obj_crowding_dist_list)
        crowding_dist = numpy.array([obj_crowding_dist_list[idx, :, 2] for idx in range(len(obj_crowding_dist_list))])
        crowding_dist_sum = numpy.sum(crowding_dist, axis=0)
    
        # An array of the sum of crowding distances across all objectives. 
        # Each row has 2 elements:
            # 1) The index of the solution.
            # 2) The sum of all crowding distances for all objective of the solution.
        crowding_dist_sum = numpy.array(list(zip(obj_crowding_dist_list[0, :, 0], crowding_dist_sum)))
        crowding_dist_sum = sorted(crowding_dist_sum, key=lambda x: x[1], reverse=True)
    
        # The sorted solutions' indices by the crowding distance.
        crowding_dist_front_sorted_indices = numpy.array(crowding_dist_sum)[:, 0]
        crowding_dist_front_sorted_indices = crowding_dist_front_sorted_indices.astype(int)
        # Note that such indices are relative to the front, NOT the population.
        # It is mandatory to map such front indices to population indices before using them to refer to the population.
        crowding_dist_pop_sorted_indices = pareto_front[:, 0]
        crowding_dist_pop_sorted_indices = crowding_dist_pop_sorted_indices[crowding_dist_front_sorted_indices]
        crowding_dist_pop_sorted_indices = crowding_dist_pop_sorted_indices.astype(int)
    
        return obj_crowding_dist_list, crowding_dist_sum, crowding_dist_front_sorted_indices, crowding_dist_pop_sorted_indices

    def sort_solutions_nsga2(self,
                             fitness,
                             find_best_solution=False):
        """
        Sort the solutions based on the fitness.
        The sorting procedure differs based on whether the problem is single-objective or multi-objective optimization.
        If it is multi-objective, then non-dominated sorting and crowding distance are applied.
        At first, non-dominated sorting is applied to classify the solutions into pareto fronts.
        Then the solutions inside each front are sorted using crowded distance.
        The solutions inside pareto front X always come before those in front X+1.
    
        Parameters
        ----------
        fitness: The fitness of the entire population.
        find_best_solution: Whether the method is called only to find the best solution or as part of the PyGAD lifecycle. This is to decide whether the pareto_fronts instance attribute is edited or not.

        Returns
        -------
        solutions_sorted : TYPE
            The indices of the sorted solutions.
    
        """
        if type(fitness[0]) in [list, tuple, numpy.ndarray]:
            # Multi-objective optimization problem.
            solutions_sorted = []
            # Split the solutions into pareto fronts using non-dominated sorting.
            pareto_fronts, solutions_fronts_indices = self.non_dominated_sorting(fitness)
            if find_best_solution:
                # Do not edit the pareto_fronts instance attribute when just getting the best solution.
                pass
            else:
                # The method is called within the regular GA lifecycle.
                # We have to edit the pareto_fronts to be assigned the latest pareto front.
                self.pareto_fronts = pareto_fronts.copy()
            for pareto_front in pareto_fronts:
                # Sort the solutions in the front using crowded distance.
                _, _, _, crowding_dist_pop_sorted_indices = self.crowding_distance(pareto_front=pareto_front.copy(),
                                                                                   fitness=fitness)
                crowding_dist_pop_sorted_indices = list(crowding_dist_pop_sorted_indices)
                # Append the sorted solutions into the list.
                solutions_sorted.extend(crowding_dist_pop_sorted_indices)
        elif type(fitness[0]) in pygad.GA.supported_int_float_types:
            # Single-objective optimization problem.
            solutions_sorted = sorted(range(len(fitness)), key=lambda k: fitness[k])
            # Reverse the sorted solutions so that the best solution comes first.
            solutions_sorted.reverse()
        else:
            raise TypeError(f'Each element in the fitness array must be of a number of an iterable (list, tuple, numpy.ndarray). But the type {type(fitness[0])} found')

        return solutions_sorted



import pygad
import numpy

actual_num_fitness_calls_default_keep = 0
actual_num_fitness_calls_no_keep = 0
actual_num_fitness_calls_keep_elitism = 0
actual_num_fitness_calls_keep_parents = 0

num_generations = 50
sol_per_pop = 10
num_parents_mating = 5

function_inputs1 = [4,-2,3.5,5,-11,-4.7] # Function 1 inputs.
function_inputs2 = [-2,0.7,-9,1.4,3,5] # Function 2 inputs.
desired_output1 = 50 # Function 1 output.
desired_output2 = 30 # Function 2 output.

#### Define the fitness functions in the top-level of the module so that they are picklable and usable in the process-based parallel processing works.
#### If the functions are defined inside a class/method/function, they are not picklable and this error is raised: AttributeError: Can't pickle local object
#### Process-based parallel processing must have the used functions picklable.
def fitness_func_batch_multi(ga_instance, solution, solution_idx):
        f = []
        for sol in solution:
            output1 = numpy.sum(sol*function_inputs1)
            output2 = numpy.sum(sol*function_inputs2)
            fitness1 = 1.0 / (numpy.abs(output1 - desired_output1) + 0.000001)
            fitness2 = 1.0 / (numpy.abs(output2 - desired_output2) + 0.000001)
            f.append([fitness1, fitness2])
        return f

def fitness_func_no_batch_multi(ga_instance, solution, solution_idx):
        output1 = numpy.sum(solution*function_inputs1)
        output2 = numpy.sum(solution*function_inputs2)
        fitness1 = 1.0 / (numpy.abs(output1 - desired_output1) + 0.000001)
        fitness2 = 1.0 / (numpy.abs(output2 - desired_output2) + 0.000001)
        return [fitness1, fitness2]

def fitness_func_batch_single(ga_instance, solution, solution_idx):
        f = []
        for sol in solution:
            output = numpy.sum(solution*function_inputs1)
            fitness = 1.0 / (numpy.abs(output - desired_output1) + 0.000001)
            f.append(fitness)
        return f

def fitness_func_no_batch_single(ga_instance, solution, solution_idx):
        output = numpy.sum(solution*function_inputs1)
        fitness = 1.0 / (numpy.abs(output - desired_output1) + 0.000001)
        return fitness


def multi_objective_problem(keep_elitism=1, 
                            keep_parents=-1,
                            fitness_batch_size=None,
                            stop_criteria=None,
                            parent_selection_type='sss',
                            mutation_type="random",
                            mutation_percent_genes="default",
                            multi_objective=False,
                            parallel_processing=None):

    if fitness_batch_size is None or (type(fitness_batch_size) in pygad.GA.supported_int_types and fitness_batch_size == 1):
        if multi_objective == True:
            fitness_func = fitness_func_no_batch_multi
        else:
            fitness_func = fitness_func_no_batch_single
    elif (type(fitness_batch_size) in pygad.GA.supported_int_types and fitness_batch_size > 1):
        if multi_objective == True:
            fitness_func = fitness_func_batch_multi
        else:
            fitness_func = fitness_func_batch_single

    ga_optimizer = pygad.GA(num_generations=num_generations,
                            sol_per_pop=sol_per_pop,
                            num_genes=6,
                            num_parents_mating=num_parents_mating,
                            fitness_func=fitness_func,
                            fitness_batch_size=fitness_batch_size,
                            mutation_type=mutation_type,
                            mutation_percent_genes=mutation_percent_genes,
                            keep_elitism=keep_elitism,
                            keep_parents=keep_parents,
                            stop_criteria=stop_criteria,
                            parent_selection_type=parent_selection_type,
                            parallel_processing=parallel_processing,
                            suppress_warnings=True)

    ga_optimizer.run()
    
    return None

def test_number_calls_fitness_function_no_parallel_processing():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=None,
                            parallel_processing=None)

def test_number_calls_fitness_function_parallel_processing_thread_1():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=None,
                            parallel_processing=['thread', 1])

def test_number_calls_fitness_function_parallel_processing_thread_2():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=None,
                            parallel_processing=['thread', 2])

def test_number_calls_fitness_function_parallel_processing_thread_5():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=None,
                            parallel_processing=['thread', 5])

def test_number_calls_fitness_function_parallel_processing_thread_5_patch_4():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=4,
                            parallel_processing=['thread', 5])

def test_number_calls_fitness_function_parallel_processing_thread_5_patch_4_multi_objective():
    multi_objective_problem(multi_objective=True,
                            fitness_batch_size=4,
                            parallel_processing=['thread', 5])

def test_number_calls_fitness_function_parallel_processing_process_1():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=None,
                            parallel_processing=['process', 1])

def test_number_calls_fitness_function_parallel_processing_process_2():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=None,
                            parallel_processing=['process', 2])

def test_number_calls_fitness_function_parallel_processing_process_5():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=None,
                            parallel_processing=['process', 5])

def test_number_calls_fitness_function_parallel_processing_process_5_patch_4():
    multi_objective_problem(multi_objective=False,
                            fitness_batch_size=4,
                            parallel_processing=['process', 5])

def test_number_calls_fitness_function_parallel_processing_process_5_patch_4_multi_objective():
    multi_objective_problem(multi_objective=True,
                            fitness_batch_size=4,
                            parallel_processing=['process', 5])

# Stop Criteria
def test_number_calls_fitness_function_multi_objective_stop_criteria_str_single_value():
    multi_objective_problem(multi_objective=True,
                            stop_criteria='reach_10')

def test_number_calls_fitness_function_multi_objective_stop_criteria_str():
    multi_objective_problem(multi_objective=True,
                            stop_criteria='reach_10_20')

def test_number_calls_fitness_function_multi_objective_stop_criteria_str_decimal():
    multi_objective_problem(multi_objective=True,
                            stop_criteria='reach_-1.0_0.5')

def test_number_calls_fitness_function_multi_objective_stop_criteria_list():
    multi_objective_problem(multi_objective=True,
                            stop_criteria=['reach_10_20', 'reach_5_2'])

def test_number_calls_fitness_function_multi_objective_stop_criteria_list_decimal():
    multi_objective_problem(multi_objective=True,
                            stop_criteria=['reach_-1.0_0.5', 'reach_5_-2.8'])

def test_number_calls_fitness_function_single_objective_stop_criteria_str():
    multi_objective_problem(multi_objective=True,
                            stop_criteria='reach_10')

def test_number_calls_fitness_function_single_objective_stop_criteria_str_decimal():
    multi_objective_problem(multi_objective=True,
                            stop_criteria='reach_-1.7')

def test_number_calls_fitness_function_single_objective_stop_criteria_list():
    multi_objective_problem(multi_objective=True,
                            stop_criteria=['reach_10', 'reach_5'])

def test_number_calls_fitness_function_single_objective_stop_criteria_list_decimal():
    multi_objective_problem(multi_objective=True,
                            stop_criteria=['reach_-1.5', 'reach_-2.8'])

if __name__ == "__main__":
    print()
    test_number_calls_fitness_function_no_parallel_processing()
    print()

    #### Thread-based Parallel Processing
    test_number_calls_fitness_function_parallel_processing_thread_1()
    print()
    test_number_calls_fitness_function_parallel_processing_thread_2()
    print()
    test_number_calls_fitness_function_parallel_processing_thread_5()
    print()
    test_number_calls_fitness_function_parallel_processing_thread_5_patch_4()
    print()
    test_number_calls_fitness_function_parallel_processing_thread_5_patch_4_multi_objective()
    print()

    #### Thread-based Parallel Processing
    test_number_calls_fitness_function_parallel_processing_process_1()
    print()
    test_number_calls_fitness_function_parallel_processing_process_2()
    print()
    test_number_calls_fitness_function_parallel_processing_process_5()
    print()
    test_number_calls_fitness_function_parallel_processing_process_5_patch_4()
    print()
    test_number_calls_fitness_function_parallel_processing_process_5_patch_4_multi_objective()
    print()

    #### Multi-Objective Stop Criteria
    test_number_calls_fitness_function_multi_objective_stop_criteria_str_single_value()
    print()
    test_number_calls_fitness_function_multi_objective_stop_criteria_str()
    print()
    test_number_calls_fitness_function_multi_objective_stop_criteria_str_decimal()
    print()
    test_number_calls_fitness_function_multi_objective_stop_criteria_list()
    print()
    test_number_calls_fitness_function_multi_objective_stop_criteria_list_decimal()
    print()

    #### Single-Objective Stop Criteria
    test_number_calls_fitness_function_single_objective_stop_criteria_str()
    print()
    test_number_calls_fitness_function_single_objective_stop_criteria_str_decimal()
    print()
    test_number_calls_fitness_function_single_objective_stop_criteria_list()
    print()
    test_number_calls_fitness_function_single_objective_stop_criteria_list_decimal()
    print()




import copy
import numpy
import tensorflow.keras

def model_weights_as_vector(model):
    """
    Reshapes the Keras model weight as a vector.

    Parameters
    ----------
    model : TYPE
        The Keras model.

    Returns
    -------
    TYPE
        The weights as a 1D vector.

    """
    weights_vector = []

    for layer in model.layers: # model.get_weights():
        if layer.trainable:
            layer_weights = layer.get_weights()
            for l_weights in layer_weights:
                vector = numpy.reshape(l_weights, newshape=(l_weights.size))
                weights_vector.extend(vector)

    return numpy.array(weights_vector)

def model_weights_as_matrix(model, weights_vector):
    """
    Reshapes the PyGAD 1D solution as a Keras weight matrix.

    Parameters
    ----------
    model : TYPE
        The Keras model.
    weights_vector : TYPE
        The PyGAD solution as a 1D vector.

    Returns
    -------
    weights_matrix : TYPE
        The Keras weights as a matrix.

    """
    weights_matrix = []

    start = 0
    for layer_idx, layer in enumerate(model.layers): # model.get_weights():
    # for w_matrix in model.get_weights():
        layer_weights = layer.get_weights()
        if layer.trainable:
            for l_weights in layer_weights:
                layer_weights_shape = l_weights.shape
                layer_weights_size = l_weights.size
        
                layer_weights_vector = weights_vector[start:start + layer_weights_size]
                layer_weights_matrix = numpy.reshape(layer_weights_vector, newshape=(layer_weights_shape))
                weights_matrix.append(layer_weights_matrix)
        
                start = start + layer_weights_size
        else:
            for l_weights in layer_weights:
                weights_matrix.append(l_weights)

    return weights_matrix

def predict(model, 
            solution, 
            data, 
            batch_size=None,
            verbose=0,
            steps=None):
    """
    Use the PyGAD's solution to make predictions using the Keras model.

    Parameters
    ----------
    model : TYPE
        The Keras model.
    solution : TYPE
        A single PyGAD solution as 1D vector.
    data : TYPE
        The data or a generator.
    batch_size : TYPE, optional
        The batch size (i.e. number of samples per step or batch). The default is None. Check documentation of the Keras Model.predict() method for more information.
    verbose : TYPE, optional
        Verbosity mode. The default is 0. Check documentation of the Keras Model.predict() method for more information.
    steps : TYPE, optional
        The total number of steps (batches of samples). The default is None. Check documentation of the Keras Model.predict() method for more information.

    Returns
    -------
    predictions : TYPE
        The Keras model predictions.

    """
    # Fetch the parameters of the best solution.
    solution_weights = model_weights_as_matrix(model=model,
                                               weights_vector=solution)
    _model = tensorflow.keras.models.clone_model(model)
    _model.set_weights(solution_weights)
    predictions = _model.predict(x=data,
                                 batch_size=batch_size,
                                 verbose=verbose,
                                 steps=steps)

    return predictions

class KerasGA:

    def __init__(self, model, num_solutions):

        """
        Creates an instance of the KerasGA class to build a population of model parameters.

        model: A Keras model class.
        num_solutions: Number of solutions in the population. Each solution has different model parameters.
        """
        
        self.model = model

        self.num_solutions = num_solutions

        # A list holding references to all the solutions (i.e. networks) used in the population.
        self.population_weights = self.create_population()

    def create_population(self):

        """
        Creates the initial population of the genetic algorithm as a list of networks' weights (i.e. solutions). Each element in the list holds a different weights of the Keras model.

        The method returns a list holding the weights of all solutions.
        """

        model_weights_vector = model_weights_as_vector(model=self.model)

        net_population_weights = []
        net_population_weights.append(model_weights_vector)

        for idx in range(self.num_solutions-1):

            net_weights = copy.deepcopy(model_weights_vector)
            net_weights = numpy.array(net_weights) + numpy.random.uniform(low=-1.0, high=1.0, size=model_weights_vector.size)

            # Appending the weights to the population.
            net_population_weights.append(net_weights)

        return net_population_weights
